@@ -334,6 +334,8 @@ struct grspw_txbd {
 /* GRSPW Constants */
 #define GRSPW_TXBD_NR 64	/* Maximum number of TX Descriptors */
 #define GRSPW_RXBD_NR 128	/* Maximum number of RX Descriptors */
+#define GRSPW_TXBD_SIZE 16	/* Size in bytes of one TX descriptor */
+#define GRSPW_RXBD_SIZE 8	/* Size in bytes of one RX descriptor */
 #define BDTAB_SIZE 0x400	/* BD Table Size (RX or TX) */
 #define BDTAB_ALIGN 0x400	/* BD Table Alignment Requirement */
 
@@ -374,7 +376,8 @@ struct grspw_dma_priv {
 	int index;			/* DMA Channel Index @ GRSPW core */
 	int open;			/* DMA Channel opened by user */
 	int started;			/* DMA Channel activity (start|stop) */
-	rtems_id sem_dma;		/* DMA Channel Semaphore */
+	rtems_id sem_rxdma;		/* DMA Channel RX Semaphore */
+	rtems_id sem_txdma;		/* DMA Channel TX Semaphore */
 	struct grspw_dma_stats stats;	/* DMA Channel Statistics */
 	struct grspw_dma_config cfg;	/* DMA Channel Configuration */
 
@@ -469,8 +472,11 @@ struct grspw_priv {
 	spwpkt_ic_isr_t icisr;
 	void *icisr_arg;
 
-	/* Disable Link on SpW Link error */
-	int dis_link_on_err;
+	/* Bit mask representing events which shall cause link disable. */
+	unsigned int dis_link_on_err;
+
+	/* Bit mask for link status bits to clear by ISR */
+	unsigned int stscfg;
 
 	/* "Core Global" Statistics gathered, not dependent on DMA channel */
 	struct grspw_core_stats stats;
@@ -538,6 +544,7 @@ void *grspw_open(int dev_no)
 	priv->tcisr_arg = NULL;
 	priv->icisr = NULL;
 	priv->icisr_arg = NULL;
+	priv->stscfg = LINKSTS_MASK;
 
 	grspw_stats_clr(priv);
 
@@ -545,6 +552,7 @@ void *grspw_open(int dev_no)
 	 * channels. Max-size descriptor area is allocated (or user assigned):
 	 *  - 128 RX descriptors per DMA Channel
 	 *  - 64 TX descriptors per DMA Channel
+	 * Specified address must be in CPU RAM.
  	 */
 	bdtabsize = 2 * BDTAB_SIZE * priv->hwsup.ndma_chans;
 	value = drvmgr_dev_key_get(priv->dev, "bdDmaArea", KEY_TYPE_INT);
@@ -618,7 +626,7 @@ out:
 	return priv;
 }
 
-void grspw_close(void *d)
+int grspw_close(void *d)
 {
 	struct grspw_priv *priv = d;
 	int i;
@@ -626,21 +634,24 @@ void grspw_close(void *d)
 	/* Take GRSPW lock - Wait until we get semaphore */
 	if (rtems_semaphore_obtain(grspw_sem, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
-		return;
+		return -1;
 
-	/* Stop Hardware from doing DMA, put HW into "startup-state",
-	 * Stop hardware from generating IRQ.
+	/* Check that user has stopped and closed all DMA channels
+	 * appropriately. At this point the Hardware shall not be doing DMA
+	 * or generating Interrupts. We want HW in a "startup-state".
 	 */
-	for (i=0; i<priv->hwsup.ndma_chans; i++)
-		grspw_dma_close(&priv->dma[i]);
+	for (i=0; i<priv->hwsup.ndma_chans; i++) {
+		if (priv->dma[i].open) {
+			rtems_semaphore_release(grspw_sem);
+			return 1;
+		}
+	}
 	grspw_hw_stop(priv);
 
 	/* Mark not open */
 	priv->open = 0;
-
 	rtems_semaphore_release(grspw_sem);
-
-	/* Check that all threads are out? */
+	return 0;
 }
 
 void grspw_hw_support(void *d, struct grspw_hw_sup *hw)
@@ -718,6 +729,14 @@ unsigned int grspw_link_status(void *d)
 	return REG_READ(&priv->regs->status);
 }
 
+/* Clear Status Register bits */
+void grspw_link_status_clr(void *d, unsigned int mask)
+{
+	struct grspw_priv *priv = d;
+
+	REG_WRITE(&priv->regs->status, mask);
+}
+
 /* Return Current Link State */
 spw_link_state_t grspw_link_state(void *d)
 {
@@ -736,7 +755,7 @@ static inline int grspw_is_irqsource_set(unsigned int ctrl, unsigned int icctrl)
 
 
 /* options and clkdiv [in/out]: set to -1 to only read current config */
-void grspw_link_ctrl(void *d, int *options, int *clkdiv)
+void grspw_link_ctrl(void *d, int *options, int *stscfg, int *clkdiv)
 {
 	struct grspw_priv *priv = d;
 	struct grspw_regs *regs = priv->regs;
@@ -763,10 +782,21 @@ void grspw_link_ctrl(void *d, int *options, int *clkdiv)
 				ctrl &= ~GRSPW_CTRL_IE;
 
 			REG_WRITE(&regs->ctrl, ctrl);
-			priv->dis_link_on_err = (*options & LINKOPTS_DIS_ONERR) >> 3;
+			/* Store the link disable events for use in
+			ISR. The LINKOPTS_DIS_ON_* options are actually the
+			corresponding bits in the status register, shifted
+			by 16. */
+			priv->dis_link_on_err = *options &
+				(LINKOPTS_MASK_DIS_ON | LINKOPTS_DIS_ONERR);
 		}
 		SPIN_UNLOCK_IRQ(&priv->devlock, irqflags);
-		*options = (ctrl & GRSPW_LINK_CFG)|(priv->dis_link_on_err << 3);
+		*options = (ctrl & GRSPW_LINK_CFG) | priv->dis_link_on_err;
+	}
+	if (stscfg) {
+		if (*stscfg != -1) {
+			priv->stscfg = *stscfg & LINKSTS_MASK;
+		}
+		*stscfg = priv->stscfg;
 	}
 }
 
@@ -827,14 +857,14 @@ void grspw_tc_isr(void *d, void (*tcisr)(void *data, int tc), void *data)
  */
 void grspw_tc_time(void *d, int *time)
 {
-        struct grspw_priv *priv = d;
-        struct grspw_regs *regs = priv->regs;
+	struct grspw_priv *priv = d;
+	struct grspw_regs *regs = priv->regs;
 
-        if (time == NULL)
-                return;
-        if (*time != -1)
-                REG_WRITE(&regs->time, *time & (GRSPW_TIME_TCNT | GRSPW_TIME_CTRL));
-        *time = REG_READ(&regs->time) & (GRSPW_TIME_TCNT | GRSPW_TIME_CTRL);
+	if (time == NULL)
+		return;
+	if (*time != -1)
+		REG_WRITE(&regs->time, *time & (GRSPW_TIME_TCNT | GRSPW_TIME_CTRL));
+	*time = REG_READ(&regs->time) & (GRSPW_TIME_TCNT | GRSPW_TIME_CTRL);
 }
 
 /* Generate Tick-In for the given Interrupt-code and check for generation
@@ -1599,7 +1629,7 @@ void *grspw_dma_open(void *d, int chan_no)
 	struct grspw_dma_priv *dma;
 	int size;
 
-	if ((chan_no < 0) && (priv->hwsup.ndma_chans <= chan_no))
+	if ((chan_no < 0) || (priv->hwsup.ndma_chans <= chan_no))
 		return NULL;
 
 	dma = &priv->dma[chan_no];
@@ -1627,24 +1657,37 @@ void *grspw_dma_open(void *d, int chan_no)
 	dma->cfg.tx_irq_en_cnt = 0;
 	dma->cfg.flags = DMAFLAG_NO_SPILL;
 
+	/* set to NULL so that error exit works correctly */
+	dma->sem_rxdma = RTEMS_ID_NONE;
+	dma->sem_txdma = RTEMS_ID_NONE;
+	dma->rx_wait.sem_wait = RTEMS_ID_NONE;
+	dma->tx_wait.sem_wait = RTEMS_ID_NONE;
+	dma->rx_ring_base = NULL;
+
 	/* DMA Channel Semaphore created with count = 1 */
 	if (rtems_semaphore_create(
-	    rtems_build_name('S', 'D', '0' + priv->index, '0' + chan_no), 1,
+	    rtems_build_name('S', 'D', '0' + priv->index, '0' + chan_no*2), 1,
 	    RTEMS_FIFO | RTEMS_SIMPLE_BINARY_SEMAPHORE | \
 	    RTEMS_NO_INHERIT_PRIORITY | RTEMS_LOCAL | \
-	    RTEMS_NO_PRIORITY_CEILING, 0, &dma->sem_dma) != RTEMS_SUCCESSFUL) {
-		dma = NULL;
-		goto out;
+	    RTEMS_NO_PRIORITY_CEILING, 0, &dma->sem_rxdma) != RTEMS_SUCCESSFUL) {
+		dma->sem_rxdma = RTEMS_ID_NONE;
+		goto err;
+	}
+	if (rtems_semaphore_create(
+	    rtems_build_name('S', 'D', '0' + priv->index, '0' + chan_no*2+1), 1,
+	    RTEMS_FIFO | RTEMS_SIMPLE_BINARY_SEMAPHORE | \
+	    RTEMS_NO_INHERIT_PRIORITY | RTEMS_LOCAL | \
+	    RTEMS_NO_PRIORITY_CEILING, 0, &dma->sem_txdma) != RTEMS_SUCCESSFUL) {
+		dma->sem_txdma = RTEMS_ID_NONE;
+		goto err;
 	}
 
 	/* Allocate memory for the two descriptor rings */
 	size = sizeof(struct grspw_ring) * (GRSPW_RXBD_NR + GRSPW_TXBD_NR);
 	dma->rx_ring_base = (struct grspw_rxring *)malloc(size);
 	dma->tx_ring_base = (struct grspw_txring *)&dma->rx_ring_base[GRSPW_RXBD_NR];
-	if (dma->rx_ring_base == NULL) {
-		dma = NULL;
-		goto out;
-	}
+	if (dma->rx_ring_base == NULL)
+		goto err;
 
 	/* Create DMA RX and TX Channel sempahore with count = 0 */
 	if (rtems_semaphore_create(
@@ -1652,16 +1695,16 @@ void *grspw_dma_open(void *d, int chan_no)
 	    RTEMS_FIFO | RTEMS_SIMPLE_BINARY_SEMAPHORE | \
 	    RTEMS_NO_INHERIT_PRIORITY | RTEMS_LOCAL | \
 	    RTEMS_NO_PRIORITY_CEILING, 0, &dma->rx_wait.sem_wait) != RTEMS_SUCCESSFUL) {
-		dma = NULL;
-		goto out;
+		dma->rx_wait.sem_wait = RTEMS_ID_NONE;
+		goto err;
 	}
 	if (rtems_semaphore_create(
 	    rtems_build_name('S', 'T', '0' + priv->index, '0' + chan_no), 0,
 	    RTEMS_FIFO | RTEMS_SIMPLE_BINARY_SEMAPHORE | \
 	    RTEMS_NO_INHERIT_PRIORITY | RTEMS_LOCAL | \
 	    RTEMS_NO_PRIORITY_CEILING, 0, &dma->tx_wait.sem_wait) != RTEMS_SUCCESSFUL) {
-		dma = NULL;
-		goto out;
+		dma->tx_wait.sem_wait = RTEMS_ID_NONE;
+		goto err;
 	}
 
 	/* Reset software structures */
@@ -1674,6 +1717,21 @@ out:
 	rtems_semaphore_release(grspw_sem);
 
 	return dma;
+
+	/* initialization error happended */
+err:
+	if (dma->sem_rxdma != RTEMS_ID_NONE)
+		rtems_semaphore_delete(dma->sem_rxdma);
+	if (dma->sem_txdma != RTEMS_ID_NONE)
+		rtems_semaphore_delete(dma->sem_txdma);
+	if (dma->rx_wait.sem_wait != RTEMS_ID_NONE)
+		rtems_semaphore_delete(dma->rx_wait.sem_wait);
+	if (dma->tx_wait.sem_wait != RTEMS_ID_NONE)
+		rtems_semaphore_delete(dma->tx_wait.sem_wait);
+	if (dma->rx_ring_base)
+		free(dma->rx_ring_base);
+	dma = NULL;
+	goto out;
 }
 
 /* Initialize Software Structures:
@@ -1710,24 +1768,40 @@ STATIC void grspw_dma_reset(struct grspw_dma_priv *dma)
 	grspw_dma_stats_clr(dma);
 }
 
-void grspw_dma_close(void *c)
+int grspw_dma_close(void *c)
 {
 	struct grspw_dma_priv *dma = c;
 
 	if (!dma->open)
-		return;
+		return 0;
 
 	/* Take device lock - Wait until we get semaphore */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
-		return;
+		return -1;
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	    != RTEMS_SUCCESSFUL) {
+	    	rtems_semaphore_release(dma->sem_rxdma);
+		return -1;
+	}
 
-	grspw_dma_stop_locked(dma);
+	/* Can not close active DMA channel. User must stop DMA and make sure
+	 * no threads are active/blocked within driver.
+	 */
+	if (dma->started || dma->rx_wait.waiting || dma->tx_wait.waiting) {
+	    	rtems_semaphore_release(dma->sem_txdma);
+		rtems_semaphore_release(dma->sem_rxdma);
+		return 1;
+	}
 
 	/* Free resources */
 	rtems_semaphore_delete(dma->rx_wait.sem_wait);
 	rtems_semaphore_delete(dma->tx_wait.sem_wait);
-	rtems_semaphore_delete(dma->sem_dma); /* Release and delete lock */
+	/* Release and delete lock. Operations requiring lock will fail */
+	rtems_semaphore_delete(dma->sem_txdma);
+	rtems_semaphore_delete(dma->sem_rxdma);
+	dma->sem_txdma = RTEMS_ID_NONE;
+	dma->sem_rxdma = RTEMS_ID_NONE;
 
 	/* Free memory */
 	if (dma->rx_ring_base)
@@ -1736,6 +1810,7 @@ void grspw_dma_close(void *c)
 	dma->tx_ring_base = NULL;
 
 	dma->open = 0;
+	return 0;
 }
 
 /* Schedule List of packets for transmission at some point in
@@ -1751,7 +1826,7 @@ int grspw_dma_tx_send(void *c, int opts, struct grspw_list *pkts, int count)
 	int ret;
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
@@ -1766,7 +1841,7 @@ int grspw_dma_tx_send(void *c, int opts, struct grspw_list *pkts, int count)
 		grspw_tx_process_scheduled(dma);
 
 	/* 2. Add the requested packets to the SEND List (USER->SEND) */
-	if (pkts) {
+	if (pkts && (count > 0)) {
 		grspw_list_append_list(&dma->send, pkts);
 		dma->send_cnt += count;
 		if (dma->stats.send_cnt_max < dma->send_cnt)
@@ -1779,7 +1854,7 @@ int grspw_dma_tx_send(void *c, int opts, struct grspw_list *pkts, int count)
 
 out:
 	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_txdma);
 
 	return ret;
 }
@@ -1791,7 +1866,7 @@ int grspw_dma_tx_reclaim(void *c, int opts, struct grspw_list *pkts, int *count)
 	int cnt, started;
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
@@ -1832,25 +1907,58 @@ int grspw_dma_tx_reclaim(void *c, int opts, struct grspw_list *pkts, int *count)
 	}
 
 	/* 3. Schedule as many packets as possible (SEND->SCHED) */
-	if ((started > 0 ) && ((opts & 2) == 0))
+	if ((started > 0) && ((opts & 2) == 0))
 		grspw_tx_schedule_send(dma);
 
 	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_txdma);
 
 	return (~started) & 1; /* signal DMA has been stopped */
 }
 
-void grspw_dma_tx_count(void *c, int *send, int *sched, int *sent)
+void grspw_dma_tx_count(void *c, int *send, int *sched, int *sent, int *hw)
 {
 	struct grspw_dma_priv *dma = c;
+	int sched_cnt, diff;
+	unsigned int hwbd;
+	struct grspw_txbd *tailbd;
+
+	/* Take device lock - Wait until we get semaphore.
+	 * The lock is taken so that the counters are in sync with each other
+	 * and that DMA descriptor table and tx_ring_tail is not being updated
+	 * during HW counter processing in this function.
+	 */
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	    != RTEMS_SUCCESSFUL)
+		return;
 
 	if (send)
 		*send = dma->send_cnt;
+	sched_cnt = dma->tx_sched_cnt;
 	if (sched)
-		*sched = dma->tx_sched_cnt;
+		*sched = sched_cnt;
 	if (sent)
 		*sent = dma->sent_cnt;
+	if (hw) {
+		/* Calculate number of descriptors (processed by HW) between
+		 * HW pointer and oldest SW pointer.
+		 */
+		hwbd = REG_READ(&dma->regs->txdesc);
+		tailbd = dma->tx_ring_tail->bd;
+		diff = ((hwbd - (unsigned int)tailbd) / GRSPW_TXBD_SIZE) &
+			(GRSPW_TXBD_NR - 1);
+		/* Handle special case when HW and SW pointers are equal
+		 * because all TX descriptors have been processed by HW.
+		 */
+		if ((diff == 0) && (sched_cnt == GRSPW_TXBD_NR) &&
+		    ((BD_READ(&tailbd->ctrl) & GRSPW_TXBD_EN) == 0)) {
+			diff = GRSPW_TXBD_NR;
+		}
+		*hw = diff;
+	}
+
+	/* Unlock DMA channel */
+	rtems_semaphore_release(dma->sem_txdma);
 }
 
 static inline int grspw_tx_wait_eval(struct grspw_dma_priv *dma)
@@ -1883,7 +1991,7 @@ static inline int grspw_tx_wait_eval(struct grspw_dma_priv *dma)
 int grspw_dma_tx_wait(void *c, int send_cnt, int op, int sent_cnt, int timeout)
 {
 	struct grspw_dma_priv *dma = c;
-	int ret, rc;
+	int ret, rc, initialized = 0;
 
 	if (timeout == 0)
 		timeout = RTEMS_NO_TIMEOUT;
@@ -1891,22 +1999,22 @@ int grspw_dma_tx_wait(void *c, int send_cnt, int op, int sent_cnt, int timeout)
 check_condition:
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
 	/* Check so that no other thread is waiting, this driver only supports
 	 * one waiter at a time.
 	 */
-	if (dma->tx_wait.waiting) {
-		ret = -1;
-		goto out;
+	if (initialized == 0 && dma->tx_wait.waiting) {
+		ret = 3;
+		goto out_release;
 	}
 
-	/* Stop if link error or similar, abort */
+	/* Stop if link error or similar (DMA stopped), abort */
 	if (dma->started == 0) {
 		ret = 1;
-		goto out;
+		goto out_release;
 	}
 
 	/* Set up Condition */
@@ -1916,10 +2024,11 @@ check_condition:
 
 	if (grspw_tx_wait_eval(dma) == 0) {
 		/* Prepare Wait */
+		initialized = 1;
 		dma->tx_wait.waiting = 1;
 
 		/* Release DMA channel lock */
-		rtems_semaphore_release(dma->sem_dma);
+		rtems_semaphore_release(dma->sem_txdma);
 
 		/* Try to take Wait lock, if this fail link may have gone down
 		 * or user stopped this DMA channel
@@ -1927,27 +2036,34 @@ check_condition:
 		rc = rtems_semaphore_obtain(dma->tx_wait.sem_wait, RTEMS_WAIT,
 						timeout);
 		if (rc == RTEMS_TIMEOUT) {
-			dma->tx_wait.waiting = 0;
-			return 2;
+			ret = 2;
+			goto out;
 		} else if (rc == RTEMS_UNSATISFIED ||
 		           rc == RTEMS_OBJECT_WAS_DELETED) {
-			dma->tx_wait.waiting = 0;
-			return 1; /* sem was flushed/deleted, means DMA stop */
-		} else if (rc != RTEMS_SUCCESSFUL)
-		    	return -1;
+			ret = 1; /* sem was flushed/deleted, means DMA stop */
+			goto out;
+		} else if (rc != RTEMS_SUCCESSFUL) {
+			/* Unknown Error */
+			ret = -1;
+			goto out;
+		} else if (dma->started == 0) {
+			ret = 1;
+			goto out;
+		}
 
 		/* Check condition once more */
 		goto check_condition;
-	} else {
-		/* No Wait needed */
-		dma->tx_wait.waiting = 0;
 	}
 
 	ret = 0;
-out:
-	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
 
+out_release:
+	/* Unlock DMA channel */
+	rtems_semaphore_release(dma->sem_txdma);
+
+out:
+	if (initialized)
+		dma->tx_wait.waiting = 0;
 	return ret;
 }
 
@@ -1958,7 +2074,7 @@ int grspw_dma_rx_recv(void *c, int opts, struct grspw_list *pkts, int *count)
 	int cnt, started;
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
@@ -2003,7 +2119,7 @@ int grspw_dma_rx_recv(void *c, int opts, struct grspw_list *pkts, int *count)
 		grspw_rx_schedule_ready(dma);
 
 	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_rxdma);
 
 	return (~started) & 1;
 }
@@ -2014,7 +2130,7 @@ int grspw_dma_rx_prepare(void *c, int opts, struct grspw_list *pkts, int count)
 	int ret;
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
@@ -2042,21 +2158,54 @@ int grspw_dma_rx_prepare(void *c, int opts, struct grspw_list *pkts, int count)
 	ret = 0;
 out:
 	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_rxdma);
 
 	return ret;
 }
 
-void grspw_dma_rx_count(void *c, int *ready, int *sched, int *recv)
+void grspw_dma_rx_count(void *c, int *ready, int *sched, int *recv, int *hw)
 {
 	struct grspw_dma_priv *dma = c;
+	int sched_cnt, diff;
+	unsigned int hwbd;
+	struct grspw_rxbd *tailbd;
+
+	/* Take device lock - Wait until we get semaphore.
+	 * The lock is taken so that the counters are in sync with each other
+	 * and that DMA descriptor table and rx_ring_tail is not being updated
+	 * during HW counter processing in this function.
+	 */
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	    != RTEMS_SUCCESSFUL)
+		return;
 
 	if (ready)
 		*ready = dma->ready_cnt;
+	sched_cnt = dma->rx_sched_cnt;
 	if (sched)
-		*sched = dma->rx_sched_cnt;
+		*sched = sched_cnt;
 	if (recv)
 		*recv = dma->recv_cnt;
+	if (hw) {
+		/* Calculate number of descriptors (processed by HW) between
+		 * HW pointer and oldest SW pointer.
+		 */
+		hwbd = REG_READ(&dma->regs->rxdesc);
+		tailbd = dma->rx_ring_tail->bd;
+		diff = ((hwbd - (unsigned int)tailbd) / GRSPW_RXBD_SIZE) &
+			(GRSPW_RXBD_NR - 1);
+		/* Handle special case when HW and SW pointers are equal
+		 * because all RX descriptors have been processed by HW.
+		 */
+		if ((diff == 0) && (sched_cnt == GRSPW_RXBD_NR) &&
+		    ((BD_READ(&tailbd->ctrl) & GRSPW_RXBD_EN) == 0)) {
+			diff = GRSPW_RXBD_NR;
+		}
+		*hw = diff;
+	}
+
+	/* Unlock DMA channel */
+	rtems_semaphore_release(dma->sem_rxdma);
 }
 
 static inline int grspw_rx_wait_eval(struct grspw_dma_priv *dma)
@@ -2089,7 +2238,7 @@ static inline int grspw_rx_wait_eval(struct grspw_dma_priv *dma)
 int grspw_dma_rx_wait(void *c, int recv_cnt, int op, int ready_cnt, int timeout)
 {
 	struct grspw_dma_priv *dma = c;
-	int ret, rc;
+	int ret, rc, initialized = 0;
 
 	if (timeout == 0)
 		timeout = RTEMS_NO_TIMEOUT;
@@ -2097,22 +2246,22 @@ int grspw_dma_rx_wait(void *c, int recv_cnt, int op, int ready_cnt, int timeout)
 check_condition:
 
 	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return -1;
 
 	/* Check so that no other thread is waiting, this driver only supports
 	 * one waiter at a time.
 	 */
-	if (dma->rx_wait.waiting) {
-		ret = -1;
-		goto out;
+	if (initialized == 0 && dma->rx_wait.waiting) {
+		ret = 3;
+		goto out_release;
 	}
 
-	/* Stop if link error or similar (MDA stopped) */
+	/* Stop if link error or similar (DMA stopped), abort */
 	if (dma->started == 0) {
 		ret = 1;
-		goto out;
+		goto out_release;
 	}
 
 	/* Set up Condition */
@@ -2122,10 +2271,11 @@ check_condition:
 
 	if (grspw_rx_wait_eval(dma) == 0) {
 		/* Prepare Wait */
+		initialized = 1;
 		dma->rx_wait.waiting = 1;
 
 		/* Release channel lock */
-		rtems_semaphore_release(dma->sem_dma);
+		rtems_semaphore_release(dma->sem_rxdma);
 
 		/* Try to take Wait lock, if this fail link may have gone down
 		 * or user stopped this DMA channel
@@ -2133,27 +2283,34 @@ check_condition:
 		rc = rtems_semaphore_obtain(dma->rx_wait.sem_wait, RTEMS_WAIT,
 		                           timeout);
 		if (rc == RTEMS_TIMEOUT) {
-			dma->rx_wait.waiting = 0;
-			return 2;
+		    	ret = 2;
+			goto out;
 		} else if (rc == RTEMS_UNSATISFIED ||
 		           rc == RTEMS_OBJECT_WAS_DELETED) {
-			dma->rx_wait.waiting = 0;
-			return 1; /* sem was flushed/deleted, means DMA stop */
-		} else if (rc != RTEMS_SUCCESSFUL)
-		    	return -1;
+			ret = 1; /* sem was flushed/deleted, means DMA stop */
+			goto out;
+		} else if (rc != RTEMS_SUCCESSFUL) {
+		    	/* Unknown Error */
+			ret = -1;
+			goto out;
+		} else if (dma->started == 0) {
+			ret = 1;
+			goto out;
+		}
 
 		/* Check condition once more */
 		goto check_condition;
-	} else {
-		/* No Wait needed */
-		dma->rx_wait.waiting = 0;
 	}
+
 	ret = 0;
 
-out:
+out_release:
 	/* Unlock DMA channel */
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_rxdma);
 
+out:
+	if (initialized)
+		dma->rx_wait.waiting = 0;
 	return ret;
 }
 
@@ -2164,7 +2321,7 @@ int grspw_dma_config(void *c, struct grspw_dma_config *cfg)
 	if (dma->started || !cfg)
 		return -1;
 
-	if (cfg->flags & ~DMAFLAG_MASK)
+	if (cfg->flags & ~(DMAFLAG_MASK | DMAFLAG2_MASK))
 		return -1;
 
 	/* Update Configuration */
@@ -2209,6 +2366,7 @@ int grspw_dma_start(void *c)
 	struct grspw_dma_priv *dma = c;
 	struct grspw_dma_regs *dregs = dma->regs;
 	unsigned int ctrl;
+	IRQFLAGS_TYPE irqflags;
 
 	if (dma->started)
 		return 0;
@@ -2258,13 +2416,16 @@ int grspw_dma_start(void *c)
 	ctrl =  GRSPW_DMACTRL_AI | GRSPW_DMACTRL_PS | GRSPW_DMACTRL_PR |
 		GRSPW_DMACTRL_TA | GRSPW_DMACTRL_RA | GRSPW_DMACTRL_RE |
 		(dma->cfg.flags & DMAFLAG_MASK) << GRSPW_DMACTRL_NS_BIT;
-	if (dma->core->dis_link_on_err)
+	if (dma->core->dis_link_on_err & LINKOPTS_DIS_ONERR)
 		ctrl |= GRSPW_DMACTRL_LE;
-	if (dma->cfg.rx_irq_en_cnt != 0)
+	if (dma->cfg.rx_irq_en_cnt != 0 || dma->cfg.flags & DMAFLAG2_RXIE)
 		ctrl |= GRSPW_DMACTRL_RI;
-	if (dma->cfg.tx_irq_en_cnt != 0)
+	if (dma->cfg.tx_irq_en_cnt != 0 || dma->cfg.flags & DMAFLAG2_TXIE)
 		ctrl |= GRSPW_DMACTRL_TI;
+	SPIN_LOCK_IRQ(&dma->core->devlock, irqflags);
+	ctrl |= REG_READ(&dma->regs->ctrl) & GRSPW_DMACTRL_EN;
 	REG_WRITE(&dregs->ctrl, ctrl);
+	SPIN_UNLOCK_IRQ(&dma->core->devlock, irqflags);
 
 	dma->started = 1; /* open up other DMA interfaces */
 
@@ -2335,14 +2496,24 @@ void grspw_dma_stop(void *c)
 {
 	struct grspw_dma_priv *dma = c;
 
+	/* If DMA channel is closed we should not access the semaphore */
+	if (!dma->open)
+		return;
+
 	/* Take DMA Channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
 	    != RTEMS_SUCCESSFUL)
 		return;
+	if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+	    != RTEMS_SUCCESSFUL) {
+		rtems_semaphore_release(dma->sem_rxdma);
+		return;
+	}
 
 	grspw_dma_stop_locked(dma);
 
-	rtems_semaphore_release(dma->sem_dma);
+	rtems_semaphore_release(dma->sem_txdma);
+	rtems_semaphore_release(dma->sem_rxdma);
 }
 
 /* Do general work, invoked indirectly from ISR */
@@ -2351,8 +2522,8 @@ static void grspw_work_shutdown_func(struct grspw_priv *priv)
 	int i;
 
 	/* Link is down for some reason, and the user has configured
-	 * that we stop all DMA channels and throw out all blocked
-	 * threads.
+	 * that we stop all (open) DMA channels and throw out all their
+	 * blocked threads.
 	 */
 	for (i=0; i<priv->hwsup.ndma_chans; i++)
 		grspw_dma_stop(&priv->dma[i]);
@@ -2366,64 +2537,83 @@ static void grspw_work_dma_func(struct grspw_dma_priv *dma)
 	unsigned int ctrl;
 	IRQFLAGS_TYPE irqflags;
 
+	/* If DMA channel is closed we should not access the semaphore */
+	if (dma->open == 0)
+		return;
+
 	rx_cond_true = 0;
 	tx_cond_true = 0;
 	dma->stats.irq_cnt++;
 
-	/* Take DMA channel lock */
-	if (rtems_semaphore_obtain(dma->sem_dma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
-	    != RTEMS_SUCCESSFUL)
-		return;
-
 	/* Look at cause we were woken up and clear source */
 	SPIN_LOCK_IRQ(&priv->devlock, irqflags);
+	if (dma->started == 0) {
+		SPIN_UNLOCK_IRQ(&priv->devlock, irqflags);
+		return;
+	}
 	ctrl = REG_READ(&dma->regs->ctrl);
 
 	/* Read/Write DMA error ? */
 	if (ctrl & GRSPW_DMA_STATUS_ERROR) {
 		/* DMA error -> Stop DMA channel (both RX and TX) */
 		SPIN_UNLOCK_IRQ(&priv->devlock, irqflags);
-		grspw_dma_stop_locked(dma);
+		grspw_dma_stop(dma);
 	} else if (ctrl & (GRSPW_DMACTRL_PR | GRSPW_DMACTRL_PS)) {
 		/* DMA has finished a TX/RX packet */
 		ctrl &= ~GRSPW_DMACTRL_AT;
-		if (dma->cfg.rx_irq_en_cnt != 0)
+		if (dma->cfg.rx_irq_en_cnt != 0 ||
+		    (dma->cfg.flags & DMAFLAG2_RXIE))
 			ctrl |= GRSPW_DMACTRL_RI;
-		if (dma->cfg.tx_irq_en_cnt != 0)
+		if (dma->cfg.tx_irq_en_cnt != 0 ||
+		    (dma->cfg.flags & DMAFLAG2_TXIE))
 			ctrl |= GRSPW_DMACTRL_TI;
 		REG_WRITE(&dma->regs->ctrl, ctrl);
 		SPIN_UNLOCK_IRQ(&priv->devlock, irqflags);
 		if (ctrl & GRSPW_DMACTRL_PR) {
 			/* Do RX Work */
+
+			/* Take DMA channel RX lock */
+			if (rtems_semaphore_obtain(dma->sem_rxdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+			    != RTEMS_SUCCESSFUL)
+				return;
+
 			dma->stats.rx_work_cnt++;
 			grspw_rx_process_scheduled(dma);
-			dma->stats.rx_work_enabled += grspw_rx_schedule_ready(dma);
-			/* Check to see if condition for waking blocked USER
-		 	 * task is fullfilled.
-			 */
-			if (dma->rx_wait.waiting) {
-				rx_cond_true = grspw_rx_wait_eval(dma);
-				if (rx_cond_true)
-					dma->rx_wait.waiting = 0;
+			if (dma->started) {
+				dma->stats.rx_work_enabled +=
+					grspw_rx_schedule_ready(dma);
+				/* Check to see if condition for waking blocked
+			 	 * USER task is fullfilled.
+				 */
+				if (dma->rx_wait.waiting)
+					rx_cond_true = grspw_rx_wait_eval(dma);
 			}
+			rtems_semaphore_release(dma->sem_rxdma);
 		}
 		if (ctrl & GRSPW_DMACTRL_PS) {
 			/* Do TX Work */
+
+			/* Take DMA channel TX lock */
+			if (rtems_semaphore_obtain(dma->sem_txdma, RTEMS_WAIT, RTEMS_NO_TIMEOUT)
+			    != RTEMS_SUCCESSFUL)
+				return;
+
 			dma->stats.tx_work_cnt++;
 			grspw_tx_process_scheduled(dma);
-			dma->stats.tx_work_enabled += grspw_tx_schedule_send(dma);
-			if (dma->tx_wait.waiting) {
-				tx_cond_true = grspw_tx_wait_eval(dma);
-				if (tx_cond_true)
-					dma->tx_wait.waiting = 0;
+			if (dma->started) {
+				dma->stats.tx_work_enabled +=
+					grspw_tx_schedule_send(dma);
+				/* Check to see if condition for waking blocked
+			 	 * USER task is fullfilled.
+				 */
+				if (dma->tx_wait.waiting)
+					tx_cond_true = grspw_tx_wait_eval(dma);
 			}
+			rtems_semaphore_release(dma->sem_txdma);
 		}
 	} else
 		SPIN_UNLOCK_IRQ(&priv->devlock, irqflags);
 
-	/* Release lock */
-	rtems_semaphore_release(dma->sem_dma);
-
 	if (rx_cond_true)
 		rtems_semaphore_release(dma->rx_wait.sem_wait);
 
@@ -2476,7 +2666,8 @@ STATIC void grspw_isr(void *data)
 
 	/* Get Status from Hardware */
 	stat = REG_READ(&priv->regs->status);
-	stat_clrmsk = stat & (GRSPW_STS_TO | GRSPW_STAT_ERROR);
+	stat_clrmsk = stat & (GRSPW_STS_TO | GRSPW_STAT_ERROR) &
+			(GRSPW_STS_TO | priv->stscfg);
 
 	/* Make sure to put the timecode handling first in order to get the
 	 * smallest possible interrupt latency
@@ -2532,6 +2723,9 @@ STATIC void grspw_isr(void *data)
 		if (stat & GRSPW_STS_PE)
 			priv->stats.err_parity++;
 
+		if (stat & GRSPW_STS_DE)
+			priv->stats.err_disconnect++;
+
 		if (stat & GRSPW_STS_ER)
 			priv->stats.err_escape++;
 
@@ -2541,7 +2735,7 @@ STATIC void grspw_isr(void *data)
 		if (stat & GRSPW_STS_WE)
 			priv->stats.err_wsync++;
 
-		if (priv->dis_link_on_err) {
+		if ((priv->dis_link_on_err >> 16) & stat) {
 			/* Disable the link, no more transfers are expected
 			 * on any DMA channel.
 			 */
@@ -2910,7 +3104,7 @@ static int grspw_common_init(void)
 		if (rtems_task_start(grspw_work_task, grspw_work_func, 0) !=
 		    RTEMS_SUCCESSFUL)
 			return -1;
-	}
+}
 
 	grspw_initialized = 1;
 	return 0;
