@@ -3,14 +3,14 @@
  */
 
 /*
- *  COPYRIGHT (c) 1989-1999.
+ *  COPYRIGHT (c) 1989-2008.
  *  On-Line Applications Research Corporation (OAR).
  *
  *  The license and distribution terms for this file may be
  *  found in the file LICENSE in this distribution or at
  *  http://www.rtems.com/license/LICENSE.
  *
- *  $Id: eventflags.h,v 1.7 2005/01/28 11:07:14 ralf Exp $
+ *  $Id$
  */
 
 #ifndef _RTEMS_ITRON_EVENTFLAGS_H
@@ -46,9 +46,7 @@ ITRON_EXTERN Objects_Information  _ITRON_Eventflags_Information;
  *  This routine performs the initialization necessary for this manager.
  */
 
-void _ITRON_Eventflags_Manager_initialization(
-  uint32_t   maximum_eventflags
-);
+void _ITRON_Eventflags_Manager_initialization(void);
 
 /*
  *  XXX insert private stuff here
