@@ -6,7 +6,7 @@
  *  found in the file LICENSE in this distribution or at
  *  http://www.rtems.com/license/LICENSE.
  *
- *  $Id: ref_mpf.c,v 1.1 2007/04/17 20:32:12 joel Exp $
+ *  $Id$
  */
 
 #if HAVE_CONFIG_H
@@ -22,8 +22,8 @@
  */
 
 ER ref_mpf(
-  T_RMPF *pk_rmpf,
-  ID      mpfid
+  T_RMPF *pk_rmpf __attribute__((unused)),
+  ID      mpfid __attribute__((unused))
 )
 {
   return E_OK;
