@@ -8,14 +8,14 @@
  *
  *  Output parameters:  NONE
  *
- *  COPYRIGHT (c) 1989-1999.
+ *  COPYRIGHT (c) 1989-2009.
  *  On-Line Applications Research Corporation (OAR).
  *
  *  The license and distribution terms for this file may be
  *  found in the file LICENSE in this distribution or at
  *  http://www.rtems.com/license/LICENSE.
  *
- *  $Id: task.c,v 1.8 2003/09/04 18:53:38 joel Exp $
+ *  $Id$
  */
 
 #include "system.h"
@@ -25,17 +25,17 @@ void *Task_1_through_3(
   void *argument
 )
 {
-  int seconds;
-  int i;
-  int status;
+  unsigned int seconds;
+  int          i;
+  int          status;
 
   for ( i=0 ; i<5 ; i++ ) {
     print_current_time( "Task1: ", "" );
     status = pthread_kill( Init_id, SIGUSR1 );
-    assert( !status );
+    rtems_test_assert(  !status );
 
     seconds = sleep( 1 );
-    assert( !seconds );
+    rtems_test_assert(  !seconds );
   }
   puts( "*** END OF POSIX TEST 2 ***" );
   rtems_test_exit( 0 );
