@@ -8,14 +8,14 @@
  *
  *  Output parameters:  NONE
  *
- *  COPYRIGHT (c) 1989-1999.
+ *  COPYRIGHT (c) 1989-2009.
  *  On-Line Applications Research Corporation (OAR).
  *
  *  The license and distribution terms for this file may be
  *  found in the file LICENSE in this distribution or at
  *  http://www.rtems.com/license/LICENSE.
  *
- *  $Id: task2.c,v 1.8 2003/09/04 18:53:47 joel Exp $
+ *  $Id$
  */
 
 #include "system.h"
@@ -27,13 +27,13 @@ rtems_task_argument argument;
 
   puts( "TA2 - is beginning to run" );
   if ( argument == Argument ) {
-    status = rtems_task_wake_after( 2 * TICKS_PER_SECOND );
+    status = rtems_task_wake_after( 2 * rtems_clock_get_ticks_per_second() );
     directive_failed( status, "rtems_task_wake_after" );
   }
 
   while( FOREVER ) {
     puts( "TA2 - rtems_task_wake_after - sleep 1/2 second" );
-    status = rtems_task_wake_after( TICKS_PER_SECOND / 2 );
+    status = rtems_task_wake_after( rtems_clock_get_ticks_per_second() / 2 );
     directive_failed( status, "rtems_task_wake_after" );
   }
 }
